------------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--                             E X P _ U T I L                              --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                            $Revision: 1.137 $                            --
--                                                                          --
--        Copyright (c) 1992,1993,1994,1995 NYU, All Rights Reserved        --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT;  see file COPYING.  If not, write --
-- to the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. --
--                                                                          --
------------------------------------------------------------------------------

with Atree;    use Atree;
with Einfo;    use Einfo;
with Elists;   use Elists;
with Expander; use Expander;
with Exp_Ch7;  use Exp_Ch7;
with Itypes;   use Itypes;
with Nlists;   use Nlists;
with Nmake;    use Nmake;
with Output;   use Output;
with Sem;      use Sem;
with Sem_Eval; use Sem_Eval;
with Sem_Res;  use Sem_Res;
with Sem_Util; use Sem_Util;
with Sinfo;    use Sinfo;
with Tbuild;   use Tbuild;
with Ttypes;   use Ttypes;
with Uintp;    use Uintp;
with Urealp;   use Urealp;

package body Exp_Util is

   -----------------------
   -- Local Subprograms --
   -----------------------

   function Make_CW_Equivalent_Type
     (T : Entity_Id;
      E : Node_Id)
      return Entity_Id;
   --  T is a class-wide type entity, E is the initial expression node that
   --  constrains T in case such as: " X: T := E" or "new T'(E)"
   --  This function returns the entity of the Equivalent type and inserts
   --  on the fly the necessary declaration such as:
   --    type anon is record
   --       _parent : Root_Type (T); constrained with E discriminants (if any)
   --       Extension : String (1 .. expr to match size of E);
   --    end record;
   --
   --  This record is compatible with any object of the class of T thanks
   --  to the first field and has the same size as E thanks to the second.

   function Make_Literal_Range
     (Loc         : Source_Ptr;
      Literal_Typ : Entity_Id;
      Index_Typ   : Entity_Id)
      return        Node_Id;
   --  Produce a Range node whose bounds are:
   --    Index_Typ'first .. Index_Typ'First + Length (Literal_Typ)
   --  this is used for expanding declarations like X : String := "sdfgdfg";

   function Make_Subtype_From_Expr
     (E       : Node_Id;
      Unc_Typ : Entity_Id)
      return    Node_Id;
   --  Creates an appropriate Subtype Indication for unconstrained object
   --  declarations. Unc_Typ can be an unconstrained array or record, or
   --  a classwide type.

   function New_Class_Wide_Subtype
     (CW_Typ : Entity_Id;
      N      : Node_Id)
      return   Entity_Id;
   --  Create an implicit subtype of CW_Typ attached to node N.

   -----------------------
   -- Package Multi_Use --
   -----------------------

   package body Multi_Use is

      -------------
      -- Prepare --
      -------------

      procedure Prepare
        (Exp      : Node_Id;
         Res      : out Exp_Id;
         Code     : out List_Id;
         Name_Req : Boolean := False)
      is
         Loc          : constant Source_Ptr := Sloc (Exp);
         Exp_Type     : constant Entity_Id  := Etype (Exp);
         Kind         : constant Node_Kind  := Nkind (Exp);
         Def_Id       : Entity_Id;
         Ref_Type     : Entity_Id;
         Ptr_Typ_Decl : Node_Id;
         New_Exp      : Node_Id;
         E            : Node_Id;
         Has_Prefix   : Boolean;

         function OK_To_Copy (N : Node_Id) return Boolean;
         --  Determines if the tree N represents an expession that can
         --  safely be copied (and for which hence no temporary is needed)

         function OK_To_Copy (L : List_Id) return Boolean;
         --  Determines if all elements of the list L can be safely copied.

         function OK_To_Copy (N : Node_Id) return Boolean is
            K : constant Node_Kind := Nkind (N);

         begin
            --  No node can be copied that has Do_Range_Check set (because
            --  the back end does not properly handle CSE on such nodes well).
            --  A similar comment applies to other error check cases below.

            if Do_Range_Check (N) then
               return False;

            --  An attribute reference can be copied unless it has a
            --  non-copyable expression or if it has Do_Overflow_Check set

            elsif K = N_Attribute_Reference then
               return not Do_Overflow_Check (N)
                 and then OK_To_Copy (Expressions (N));

            --  An entity can be copied unless it is a function call, or a
            --  reference to a volatile variable and Name_Req is False. If
            --  Name_Req is True then we can't help returning a name which
            --  effectively allows multiple references in any case.

            elsif Is_Entity_Name (N)
              and then (Ekind (Entity (N)) /= E_Function)
              and then (not Is_Volatile (Entity (N)) or else Name_Req)
            then
               return True;

            --  A static expression can always be copied. This catches
            --  the cases of literals.

            elsif Is_OK_Static_Expression (N) then
               return True;

            --  A type conversion can be copied if the expression involved
            --  can itself be copied, and if Do_Overflow_Check is not set.

            elsif K = N_Type_Conversion then
               return not Do_Overflow_Check (N)
                 and then OK_To_Copy (Expression (N));

            --  An unchecked type conversion or type qualification can be
            --  copied if the expression involved can itself be copied.

            elsif K = N_Qualified_Expression
              or else K = N_Unchecked_Type_Conversion
            then
               return OK_To_Copy (Expression (N));

            --  A unary operator can be copied if its argument is copyable
            --  and the Do_Overflow_Check flag is not set.

            elsif K in N_Unary_Op then
               return not Do_Overflow_Check (N)
                 and then OK_To_Copy (Right_Opnd (N));

            --  A binary operator can be copied if its arguments are copyable
            --  and the Do_Overflow_Check flag is not set

            elsif K in N_Binary_Op then
               return not Do_Overflow_Check (N)
                 and then OK_To_Copy (Left_Opnd  (N))
                 and then OK_To_Copy (Right_Opnd (N));

            --  An explicit dererence or selected component can be copied
            --  if its prefix is copyable.

            elsif K = N_Explicit_Dereference
              or else K = N_Selected_Component
            then
               return OK_To_Copy (Prefix (N));

            --  An indexed component can be copied if the prefix is copyable
            --  and all the indexing expressions are copyable and there is
            --  no access check

            elsif K = N_Indexed_Component then
               return not Do_Access_Check (N)
                 and then OK_To_Copy (Prefix (N))
                 and then OK_To_Copy (Expressions (N));

            --  Anything else, we do not copy, we are not catching all cases
            --  but near enough!

            else
               return False;
            end if;
         end OK_To_Copy;

         function OK_To_Copy (L : List_Id) return Boolean is
            N : Node_Id;

         begin
            if L = No_List or else L = Error_List then
               return True;

            else
               N := First (L);

               while Present (N) loop
                  if not OK_To_Copy (N) then
                     return False;
                  else
                     N := Next (N);
                  end if;
               end loop;

               return True;
            end if;
         end OK_To_Copy;

      --  Start of processing for Prepare

      begin
         --  First we check for the cases for which we simply allow copies

         if OK_To_Copy (Exp) then
            Res := Exp_Id (Exp);
            Code := No_List;
            return;

         --  If the expression has the form v.all then we already have a
         --  pointer to the expression, we don't need to create one.

         elsif Nkind (Exp) = N_Explicit_Dereference
           and then OK_To_Copy (Prefix (Exp))
         then
            Res := Exp_Id (Make_Explicit_Dereference (Loc, Prefix (Exp)));
            Code := No_List;

         --  For expressions that denote objects, we can use a renaming scheme
         --  We skip using this if we have a volatile variable and we do
         --  not have Nam_Req set true (see comments above for OK_To_Copy).

         elsif Is_Object_Reference (Exp)
           and then (Name_Req
                      or else not Is_Entity_Name (Exp)
                      or else not Is_Volatile (Entity (Exp)))
         then
            Code := New_List;
            Def_Id := Make_Defining_Identifier (Loc, New_Internal_Name ('R'));
            Res := Exp_Id (New_Reference_To (Def_Id, Loc));

            Append_To (Code,
              Make_Object_Renaming_Declaration (Loc,
                Defining_Identifier => Def_Id,
                Subtype_Mark        => New_Reference_To (Exp_Type, Loc),
                Name                => Relocate_Node (Exp)));

         --  If it is a scalar type, just make a copy

         elsif Is_Elementary_Type (Exp_Type) then
            Code := New_List;

            Def_Id := Make_Defining_Identifier (Loc, New_Internal_Name ('R'));
            Res := Exp_Id (New_Reference_To (Def_Id, Loc));

            Append_To (Code,
              Make_Object_Declaration (Loc,
                Defining_Identifier => Def_Id,
                Object_Definition   => New_Reference_To (Exp_Type, Loc),
                Expression          => Relocate_Node (Exp)));

         --  Otherwise we generate a reference to the value

         else
            Ref_Type :=
              Make_Defining_Identifier (Loc, New_Internal_Name ('A'));
            Code := New_List;

            Ptr_Typ_Decl :=
              Make_Full_Type_Declaration (Loc,
                Defining_Identifier => Ref_Type,
                Type_Definition =>
                  Make_Access_To_Object_Definition (Loc,
                    All_Present => True,
                    Subtype_Indication =>
                      New_Reference_To (Exp_Type, Loc)));

            --  If the type of the expression is an implicit type defined
            --  in the expression node itself it better be transfered to
            --  the access type declaration to avoid forward references.
            --  Same thing for expression actions appearing in the prefix

            Has_Prefix := False
              or else Kind = N_Attribute_Reference
              or else Kind = N_Expanded_Name
              or else Kind = N_Explicit_Dereference
              or else Kind = N_Indexed_Component
              or else Kind = N_Reference
              or else Kind = N_Selected_Component
              or else Kind = N_Slice;

            if Is_Itype (Exp_Type) then
               if Nkind (Exp) in N_Has_Itypes then
                  Transfer_Itypes (From => Exp, To => Ptr_Typ_Decl);
               elsif Has_Prefix
                 and then Nkind (Prefix (Exp)) in N_Has_Itypes
               then
                  Transfer_Itypes (From => Prefix (Exp), To => Ptr_Typ_Decl);
               end if;
            end if;

            --  Furthermore if the expression is an expression-actions then
            --  move actions upfront too since the type can depend on those
            --  actions

            if Nkind (Exp) = N_Expression_Actions then
               Append_List_To (Code, Actions (Exp));
               E := Expression (Exp);

            elsif Has_Prefix
              and then Nkind (Prefix (Exp)) = N_Expression_Actions
            then
               Append_List_To (Code, Actions (Prefix (Exp)));
               Set_Prefix (Exp, Expression (Prefix (Exp)));
               E := Exp;

            else
               E := Exp;
            end if;

            Append_To (Code, Ptr_Typ_Decl);

            Def_Id := Make_Defining_Identifier (Loc, New_Internal_Name ('R'));
            Res := Exp_Id (
              Make_Explicit_Dereference (Loc,
                Prefix => New_Reference_To (Def_Id, Loc)));

            if Nkind (E) = N_Explicit_Dereference then
               New_Exp := Relocate_Node (Prefix (E));
            else
               New_Exp := Make_Reference (Loc, Relocate_Node (E));
            end if;

            Append_To (Code,
              Make_Object_Declaration (Loc,
                Defining_Identifier => Def_Id,
                Object_Definition   => New_Reference_To (Ref_Type, Loc),
                Expression          => New_Exp));
         end if;

         --  Preserve the Assignment_OK flag in all copies, since at least
         --  one copy may be used in a context where this flag must be set
         --  (otherwise why would the flag be set in the first place).

         Set_Assignment_OK (Node_Id (Res), Assignment_OK (Exp));
      end Prepare;

      ----------------
      -- New_Exp_Id --
      ----------------

      procedure New_Exp_Id
        (Exp      : Node_Id;
         N        : Node_Id;
         Res      : out Exp_Id;
         Name_Req : Boolean := False)
      is
         Code : List_Id;

      begin
         Prepare (Exp, Res, Code, Name_Req);
         Insert_List_Before_And_Analyze (N, Code);
      end New_Exp_Id;


      procedure New_Exp_Id
        (Exp      : Node_Id;
         L        : List_Id;
         Res      : out Exp_Id;
         Name_Req : Boolean := False)
      is
         Code : List_Id;

      begin
         Prepare (Exp, Res, Code, Name_Req);

         if Present (Code) then
            Append_List_To (L, Code);
         end if;
      end New_Exp_Id;

      -------------
      -- New_Ref --
      -------------

      function New_Ref (E : Exp_Id) return  Node_Id is
         Cop : constant Node_Id := New_Copy_Tree (Node_Id (E));

      begin
         --  Set paren level to 1 in copy so that it gets treated as a unit
         --  always in the source tree printout.

         Set_Paren_Count (Cop, 1);
         return Cop;
      end New_Ref;

      ----------
      -- Wrap --
      ----------

      function Wrap (Code : List_Id; N : Node_Id) return Node_Id is
      begin
         if No (Code) then
            return N;
         else
            return
              Make_Expression_Actions (Sloc (N),
                Actions    => Code,
                Expression => N);
         end if;
      end Wrap;

   end Multi_Use;

   --------------------------
   -- Append_Freeze_Action --
   --------------------------

   procedure Append_Freeze_Action (T : Entity_Id; N : Node_Id) is
      Fnode : constant Node_Id := Freeze_Node (T);

   begin
      if not Present (Actions (Fnode)) then
         Set_Actions (Fnode, New_List);
      end if;

      Append (N, Actions (Fnode));
   end Append_Freeze_Action;

   ---------------------------
   -- Append_Freeze_Actions --
   ---------------------------

   procedure Append_Freeze_Actions (T : Entity_Id; L : List_Id) is
      Fnode : constant Node_Id := Freeze_Node (T);

   begin
      if No (L) then
         return;

      else
         if No (Actions (Fnode)) then
            Set_Actions (Fnode, L);

         else
            Append_List (L, Actions (Fnode));
         end if;

      end if;
   end Append_Freeze_Actions;

   ------------------------
   -- Build_Runtime_Call --
   ------------------------

   function Build_Runtime_Call (Loc : Source_Ptr; RE : RE_Id) return Node_Id is
   begin
      return
        Make_Procedure_Call_Statement (Loc,
          Name => New_Reference_To (RTE (RE), Loc));
   end Build_Runtime_Call;

   --------------------------
   -- Compile_Time_Compare --
   --------------------------

   function Compile_Time_Compare (L, R : Node_Id) return Compare_Result is
      T : constant Entity_Id := Etype (L);

   begin
      --  Static expression case

      if Is_OK_Static_Expression (L)
        and then Is_OK_Static_Expression (R)
      then
         --  For the floating-point case, we have to be a little careful, since
         --  at compile time we are dealing with universal exact values, but at
         --  runtime, these will be in non-exact target form. That's why the
         --  returned results are LE and GE below instead of LT and GT.

         if Is_Floating_Point_Type (T) then
            declare
               Lo : constant Ureal := Expr_Value_R (L);
               Hi : constant Ureal := Expr_Value_R (R);

            begin
               if Lo < Hi then
                  return LE;
               elsif Lo = Hi then
                  return EQ;
               else
                  return GE;
               end if;
            end;

         --  For the integer case we know exactly (note that this includes the
         --  fixed-point case, where we know the run time integer values now)

         else
            declare
               Lo : constant Uint := Expr_Value (L);
               Hi : constant Uint := Expr_Value (R);

            begin
               if Lo < Hi then
                  return LT;
               elsif Lo = Hi then
                  return EQ;
               else
                  return GT;
               end if;
            end;
         end if;

      --  For now, say not known if non-static expressions, we will do better
      --  than this later on for some special cases where we can tell.

      else
         return Unknown;
      end if;
   end Compile_Time_Compare;

   ----------------
   -- Convert_To --
   ----------------

   function Convert_To (Typ : Entity_Id; Expr : Node_Id) return Node_Id is
   begin
      if Present (Etype (Expr))
        and then (Etype (Expr)) = Typ
      then
         return Relocate_Node (Expr);
      else
         return
           Make_Type_Conversion (Sloc (Expr),
             Subtype_Mark => New_Occurrence_Of (Typ, Sloc (Expr)),
             Expression => Relocate_Node (Expr));
      end if;
   end Convert_To;

   -------------------------------
   -- Convert_To_Actual_Subtype --
   -------------------------------

   function Convert_To_Actual_Subtype (Exp : Entity_Id) return Entity_Id is
      Act_ST : Entity_Id;

   begin
      Act_ST := Get_Actual_Subtype (Exp);

      if Act_ST = Etype (Exp) then
         return Exp;

      else
         Rewrite_Substitute_Tree (Exp,
           Convert_To (Get_Actual_Subtype (Exp), Relocate_Node (Exp)));

         Analyze (Exp);
         Resolve (Exp, Act_ST);
         return Exp;
      end if;
   end Convert_To_Actual_Subtype;

   -----------------------
   -- Duplicate_Subexpr --
   -----------------------

   function Duplicate_Subexpr
     (Exp      : Node_Id;
      Name_Req : Boolean := False)
      return Node_Id
   is
   begin
      Remove_Side_Effects (Exp, Name_Req);
      return New_Copy_Tree (Exp);
   end Duplicate_Subexpr;

   -------------------
   -- Insert_Action --
   -------------------

   procedure Insert_Action (Assoc_Node : Node_Id; Ins_Action : Node_Id) is
   begin
      if Present (Ins_Action) then
         Insert_Actions (Assoc_Node, New_List (Ins_Action));
      end if;
   end Insert_Action;

   --  Version with check(s) suppressed

   procedure Insert_Action
     (Assoc_Node : Node_Id; Ins_Action : Node_Id; Suppress : Check_Id)
   is
   begin
      Insert_Actions (Assoc_Node, New_List (Ins_Action), Suppress);
   end Insert_Action;

   --------------------
   -- Insert_Actions --
   --------------------

   procedure Insert_Actions (Assoc_Node : Node_Id; Ins_Actions : List_Id) is
      N : Node_Id;
      P : Node_Id;

      Wrapped_Node : Node_Id;

   begin
      if No (Ins_Actions) or else Is_Empty_List (Ins_Actions) then
         return;
      end if;

      --  We now intend to climb up the tree to find the right point to
      --  insert the actions. We start at Assoc_Node, unless this node is
      --  a subexpression in which case we start with its parent. We do this
      --  for two reasons. First it speeds things up. Second, if Assoc_Node
      --  is itself one of the special nodes like N_And_Then, then we assume
      --  that an initial request to insert actions for such a node does not
      --  expect the actions to get deposited in the node for later handling
      --  when the node is expanded, since clearly the node is being dealt
      --  with by the caller. Note that in the subexpression case, N is
      --  always the child we came from.

      if Nkind (Assoc_Node) in N_Subexpr then
         P := Assoc_Node;
         N := Parent (Assoc_Node);

      --  Non-subexpression case. Note that N is initially undefined in
      --  this case (N is only guaranteed defined in the subexpr case).

      else
         P := Assoc_Node;
      end if;

      --  Capture root of the transient scope

      if Scope_Is_Transient then
         Wrapped_Node  := Node_To_Be_Wrapped;
      else
         Wrapped_Node := Empty;
      end if;

      loop
         pragma Assert (Present (P));

         --  Make sure that inserted actions stay in the transient scope

         if P = Wrapped_Node then
            Store_Actions_In_Scope (Ins_Actions);
            return;
         end if;

         case Nkind (P) is

            --  Case of right operand of AND THEN or OR ELSE. Put the actions
            --  in the Actions field of the right operand. They will be moved
            --  out further when the AND THEN or OR ELSE operator is expanded.
            --  Nothing special needs to be done for the left operand since
            --  in that case the actions are executed unconditionally.

            when N_And_Then | N_Or_Else =>
               if N = Right_Opnd (P) then
                  if Present (Actions (P)) then
                     Insert_List_After_And_Analyze
                      (Last (Actions (P)), Ins_Actions);
                  else
                     Set_Actions (P, Ins_Actions);
                     Analyze_List (Actions (P));
                  end if;

                  return;
               end if;

            --  Then or Else operand of conditional expression. Add actions to
            --  Then_Actions or Else_Actions field as appropriate. The actions
            --  will be moved further out when the conditional is expanded.

            when N_Conditional_Expression =>
               declare
                  ThenX : constant Node_Id := Next (First (Expressions (N)));
                  ElseX : constant Node_Id := Next (ThenX);

               begin
                  --  Actions belong to the then expression, temporarily
                  --  place them as Then_Actions of the conditional expr.
                  --  They will be moved to the proper place later when
                  --  the conditional expression is expanded.

                  if N = ThenX then
                     if Present (Then_Actions (P)) then
                        Insert_List_After_And_Analyze
                          (Last (Then_Actions (P)), Ins_Actions);
                     else
                        Set_Then_Actions (P, Ins_Actions);
                        Analyze_List (Then_Actions (P));
                     end if;

                     return;

                  --  Actions belong to the else expression, temporarily
                  --  place them as Else_Actions of the conditional expr.
                  --  They will be moved to the proper place later when
                  --  the conditional expression is expanded.

                  elsif N = ElseX then
                     if Present (Else_Actions (P)) then
                        Insert_List_After_And_Analyze
                          (Last (Else_Actions (P)), Ins_Actions);
                     else
                        Set_Else_Actions (P, Ins_Actions);
                        Analyze_List (Else_Actions (P));
                     end if;

                     return;

                  --  Actions belong to the condition. In this case they are
                  --  unconditionally executed, and so we can continue the
                  --  search for the proper insert point.

                  else
                     null;
                  end if;
               end;

            --  Case of appearing in the condition of a while expression or
            --  elsif. We insert the actions into the Condition_Actions field.
            --  They will be moved further out when the while loop or elsif
            --  is anaylyzed.

            when N_Iteration_Scheme |
                 N_Elsif_Part
            =>
               if N = Condition (P) then
                  if Present (Condition_Actions (P)) then
                     Insert_List_After_And_Analyze
                       (Last (Condition_Actions (P)), Ins_Actions);
                  else
                     Set_Condition_Actions (P, Ins_Actions);
                     Analyze_List (Condition_Actions (P));
                  end if;

                  return;
               end if;

            --  Statements, declarations, pragmas, representation clauses.

            when
               --  Statements

               N_Accept_Statement                       |
               N_Procedure_Call_Statement               |
               N_Statement                              |

               --  Pragmas

               N_Pragma                                 |

               --  Representation_Clause

               N_At_Clause                              |
               N_Attribute_Definition_Clause            |
               N_Enumeration_Representation_Clause      |
               N_Record_Representation_Clause           |

               --  Declarations

               N_Abstract_Subprogram_Declaration        |
               N_Entry_Body                             |
               N_Entry_Declaration                      |
               N_Exception_Declaration                  |
               N_Exception_Renaming_Declaration         |
               N_Formal_Object_Declaration              |
               N_Formal_Subprogram_Declaration          |
               N_Formal_Type_Declaration                |
               N_Full_Type_Declaration                  |
               N_Function_Instantiation                 |
               N_Generic_Function_Renaming_Declaration  |
               N_Generic_Package_Declaration            |
               N_Generic_Package_Renaming_Declaration   |
               N_Generic_Procedure_Renaming_Declaration |
               N_Generic_Subprogram_Declaration         |
               N_Implicit_Label_Declaration             |
               N_Incomplete_Type_Declaration            |
               N_Number_Declaration                     |
               N_Object_Declaration                     |
               N_Object_Renaming_Declaration            |
               N_Package_Body                           |
               N_Package_Body_Stub                      |
               N_Package_Declaration                    |
               N_Package_Instantiation                  |
               N_Package_Renaming_Declaration           |
               N_Private_Extension_Declaration          |
               N_Private_Type_Declaration               |
               N_Procedure_Instantiation                |
               N_Protected_Body_Stub                    |
               N_Protected_Type_Declaration             |
               N_Single_Task_Declaration                |
               N_Subprogram_Body                        |
               N_Subprogram_Body_Stub                   |
               N_Subprogram_Declaration                 |
               N_Subprogram_Renaming_Declaration        |
               N_Subtype_Declaration                    |
               N_Task_Body                              |
               N_Task_Body_Stub                         |
               N_Task_Type_Declaration
            =>
               --  Do not insert here if the item is not a list member (this
               --  happens for example with a triggering statement, and the
               --  proper approach is to insret before the entire select).

               if not Is_List_Member (P) then
                  null;

               --  Do not insert if parent of P is an N_Component_Association
               --  node (i.e. we are in the context of an N_Aggregate node.
               --  In this case we want to insert before the entire aggregate.

               elsif Nkind (Parent (P)) = N_Component_Association then
                  null;

               --  Do not insert if the parent of P is either an N_Variant
               --  node or an N_Record_Definition node, meaning in either
               --  case that P is a member of a component list, and that
               --  therefore the actions should be inserted outside the
               --  complete record declaration.

               elsif Nkind (Parent (P)) = N_Variant
                 or else Nkind (Parent (P)) = N_Record_Definition
               then
                  null;

               --  Otherwise we can go ahead and do the insertion

               else
                  Insert_List_Before_And_Analyze (P, Ins_Actions);
                  return;
               end if;

            --  For all other node types, keep climbing tree

            when
               N_Abortable_Part                         |
               N_Accept_Alternative                     |
               N_Access_Definition                      |
               N_Access_Function_Definition             |
               N_Access_Procedure_Definition            |
               N_Access_To_Object_Definition            |
               N_Aggregate                              |
               N_Allocator                              |
               N_Attribute_Reference                    |
               N_Case_Statement_Alternative             |
               N_Character_Literal                      |
               N_Compilation_Unit                       |
               N_Component_Association                  |
               N_Component_Clause                       |
               N_Component_Declaration                  |
               N_Component_List                         |
               N_Concat_Multiple                        |
               N_Constrained_Array_Definition           |
               N_Decimal_Fixed_Point_Definition         |
               N_Defining_Character_Literal             |
               N_Defining_Identifier                    |
               N_Defining_Operator_Symbol               |
               N_Defining_Program_Unit_Name             |
               N_Delay_Alternative                      |
               N_Delta_Constraint                       |
               N_Derived_Type_Definition                |
               N_Designator                             |
               N_Digits_Constraint                      |
               N_Discriminant_Association               |
               N_Discriminant_Specification             |
               N_Empty                                  |
               N_Entry_Body_Formal_Part                 |
               N_Entry_Call_Alternative                 |
               N_Entry_Index_Specification              |
               N_Enumeration_Type_Definition            |
               N_Error                                  |
               N_Exception_Handler                      |
               N_Expanded_Name                          |
               N_Explicit_Dereference                   |
               N_Expression_Actions                     |
               N_Extension_Aggregate                    |
               N_Floating_Point_Definition              |
               N_Formal_Decimal_Fixed_Point_Definition  |
               N_Formal_Derived_Type_Definition         |
               N_Formal_Discrete_Type_Definition        |
               N_Formal_Floating_Point_Definition       |
               N_Formal_Modular_Type_Definition         |
               N_Formal_Ordinary_Fixed_Point_Definition |
               N_Formal_Package_Declaration             |
               N_Formal_Private_Type_Definition         |
               N_Formal_Signed_Integer_Type_Definition  |
               N_Freeze_Entity                          |
               N_Function_Call                          |
               N_Function_Specification                 |
               N_Generic_Association                    |
               N_Handled_Sequence_Of_Statements         |
               N_Identifier                             |
               N_Implicit_Types                         |
               N_In                                     |
               N_Index_Or_Discriminant_Constraint       |
               N_Indexed_Component                      |
               N_Integer_Literal                        |
               N_Interpretation                         |
               N_Label                                  |
               N_Loop_Parameter_Specification           |
               N_Mod_Clause                             |
               N_Modular_Type_Definition                |
               N_Not_In                                 |
               N_Null                                   |
               N_Op_Abs                                 |
               N_Op_Add                                 |
               N_Op_And                                 |
               N_Op_Concat                              |
               N_Op_Divide                              |
               N_Op_Eq                                  |
               N_Op_Expon                               |
               N_Op_Ge                                  |
               N_Op_Gt                                  |
               N_Op_Le                                  |
               N_Op_Lt                                  |
               N_Op_Minus                               |
               N_Op_Mod                                 |
               N_Op_Multiply                            |
               N_Op_Ne                                  |
               N_Op_Not                                 |
               N_Op_Or                                  |
               N_Op_Plus                                |
               N_Op_Rem                                 |
               N_Op_Rotate_Left                         |
               N_Op_Rotate_Right                        |
               N_Op_Shift_Left                          |
               N_Op_Shift_Right                         |
               N_Op_Shift_Right_Arithmetic              |
               N_Op_Subtract                            |
               N_Op_Xor                                 |
               N_Operator_Symbol                        |
               N_Ordinary_Fixed_Point_Definition        |
               N_Others_Choice                          |
               N_Package_Specification                  |
               N_Parameter_Association                  |
               N_Parameter_Specification                |
               N_Pragma_Argument_Association            |
               N_Procedure_Specification                |
               N_Protected_Body                         |
               N_Protected_Definition                   |
               N_Qualified_Expression                   |
               N_Raise_Constraint_Error                 |
               N_Range                                  |
               N_Range_Constraint                       |
               N_Real_Literal                           |
               N_Real_Range_Specification               |
               N_Record_Definition                      |
               N_Reference                              |
               N_Selected_Component                     |
               N_Signed_Integer_Type_Definition         |
               N_Single_Protected_Declaration           |
               N_Slice                                  |
               N_String_Literal                         |
               N_Subtype_Indication                     |
               N_Subunit                                |
               N_Task_Definition                        |
               N_Terminate_Alternative                  |
               N_Triggering_Alternative                 |
               N_Type_Conversion                        |
               N_Unchecked_Type_Conversion              |
               N_Unconstrained_Array_Definition         |
               N_Unused_At_End                          |
               N_Unused_At_Start                        |
               N_Use_Package_Clause                     |
               N_Use_Type_Clause                        |
               N_Variant                                |
               N_Variant_Part                           |
               N_With_Clause
            =>
               null;

         end case;


         --  If we fall through above tests, keep climbing tree

         N := P;
         P := Parent (N);
      end loop;

   end Insert_Actions;

   --  Version with check(s) suppressed

   procedure Insert_Actions
     (Assoc_Node : Node_Id; Ins_Actions : List_Id; Suppress : Check_Id)
   is
   begin
      if Suppress = All_Checks then
         declare
            Svg : constant Suppress_Record := Scope_Suppress;

         begin
            Scope_Suppress := (others => True);
            Insert_Actions (Assoc_Node, Ins_Actions);
            Scope_Suppress := Svg;
         end;

      else
         declare
            Svg : constant Boolean := Get_Scope_Suppress (Suppress);

         begin
            Set_Scope_Suppress (Suppress, True);
            Insert_Actions (Assoc_Node, Ins_Actions);
            Set_Scope_Suppress (Suppress, Svg);
         end;
      end if;
   end Insert_Actions;

   ----------------------------
   -- New_Class_Wide_Subtype --
   ----------------------------

   function New_Class_Wide_Subtype
     (CW_Typ : Entity_Id;
      N      : Node_Id)
      return   Entity_Id
   is
      Res      : Entity_Id := New_Itype (E_Void, N);
      Res_Name : constant Name_Id := Chars (Res);

   begin
      Set_Public_Status (Res);
      Copy_Node (CW_Typ, Res);
      Set_Chars (Res, Res_Name);
      Set_Ekind (Res, E_Class_Wide_Subtype);
      Set_Next_Entity (Res, Empty);
      Set_Etype (Res, Base_Type (CW_Typ));
      Set_Freeze_Node (Res, Empty);
      Set_Next_Itype  (Res, Empty);
      return (Res);
   end New_Class_Wide_Subtype;

   -----------------------------
   -- Make_CW_Equivalent_Type --
   -----------------------------

   --  Create a record type used as an equivalent of any member
   --  of the class which takes its size from exp.

   --  Generate the following code:

   --   type Equiv_T is record
   --     _parent :  T (List of discriminant constaints taken from Exp);
   --     Ext__50 : Storage_Array (1 .. (Exp'size - Typ'size) / Storage_Unit);
   --   end Equiv_T;

   function Make_CW_Equivalent_Type
     (T : Entity_Id;
      E : Node_Id)
      return Entity_Id
   is
      Loc         : constant Source_Ptr := Sloc (E);
      Root_Typ    : constant Entity_Id := Root_Type (T);
      Equiv_Type  : Entity_Id;
      Range_Type  : Entity_Id;
      Str_Type    : Entity_Id;
      List_Def    : List_Id := Empty_List;
      Constr_Root : Entity_Id;
      Sizexpr     : Node_Id;

   begin
      if not Has_Discriminants (Root_Typ) then
         Constr_Root := Root_Typ;
      else
         Constr_Root :=
           Make_Defining_Identifier (Loc, New_Internal_Name ('R'));

         --  subtype cstr__n is T (List of discr constraints taken from Exp)

         Append_To (List_Def,
           Make_Subtype_Declaration (Loc,
             Defining_Identifier => Constr_Root,
               Subtype_Indication =>
                 Make_Subtype_From_Expr (E, Root_Typ)));
      end if;

      --  subtype rg__xx is Storage_Offset range
      --                           (Expr'size - typ'size) / Storage_Unit

      Range_Type := Make_Defining_Identifier (Loc, New_Internal_Name ('G'));

      Sizexpr :=
        Make_Op_Subtract (Loc,
          Left_Opnd =>
            Make_Attribute_Reference (Loc,
              Prefix         => Duplicate_Subexpr (E),
              Attribute_Name => Name_Size),
          Right_Opnd =>
            Make_Attribute_Reference (Loc,
              Prefix => New_Reference_To (Constr_Root, Loc),
              Attribute_Name => Name_Size));

      Set_Paren_Count (Sizexpr, 1);

      Append_To (List_Def,
        Make_Subtype_Declaration (Loc,
          Defining_Identifier => Range_Type,
          Subtype_Indication =>
            Make_Subtype_Indication (Loc,
              Subtype_Mark => New_Reference_To (RTE (RE_Storage_Offset), Loc),
              Constraint => Make_Range_Constraint (Loc,
                Range_Expression =>
                  Make_Range (Loc,
                    Low_Bound => Make_Integer_Literal (Loc, Uint_1),
                    High_Bound =>
                      Make_Op_Divide (Loc,
                        Left_Opnd => Sizexpr,
                        Right_Opnd => Make_Integer_Literal (Loc,
                          Intval =>
                            UI_From_Int (System_Storage_Unit))))))));

      --  subtype str__nn is Storage_Array (rg__x);

      Str_Type := Make_Defining_Identifier (Loc, New_Internal_Name ('S'));
      Append_To (List_Def,
        Make_Subtype_Declaration (Loc,
          Defining_Identifier => Str_Type,
          Subtype_Indication =>
            Make_Subtype_Indication (Loc,
              Subtype_Mark => New_Reference_To (RTE (RE_Storage_Array), Loc),
              Constraint =>
                Make_Index_Or_Discriminant_Constraint (Loc,
                  Constraints =>
                    New_List (New_Reference_To (Range_Type, Loc))))));

      --  type Equiv_T is record
      --    _parent : Tnn;
      --    E : Str_Type;
      --  end Equiv_T;

      Equiv_Type := Make_Defining_Identifier (Loc, New_Internal_Name ('T'));

      --  Avoid the generation of an init procedure

      Set_Is_Frozen (Equiv_Type);

      Append_To (List_Def,
        Make_Full_Type_Declaration (Loc,
          Defining_Identifier => Equiv_Type,
          Type_Definition =>
            Make_Record_Definition (Loc,
              Component_List => Make_Component_List (Loc,
                Component_Items => New_List (
                  Make_Component_Declaration (Loc,
                    Defining_Identifier =>
                      Make_Defining_Identifier (Loc, Name_uParent),
                    Subtype_Indication => New_Reference_To (Constr_Root, Loc)),
                  Make_Component_Declaration (Loc,
                    Defining_Identifier =>
                      Make_Defining_Identifier (Loc,
                        Chars => New_Internal_Name ('X')),
                    Subtype_Indication => New_Reference_To (Str_Type, Loc))),
                Variant_Part => Empty))));

      Insert_Actions (E, List_Def);
      return Equiv_Type;
   end Make_CW_Equivalent_Type;

   ------------------------------
   -- Expand_Subtype_From_Expr --
   ------------------------------

   --  This function is applicable for both static and dynamic allocation of
   --  objects which are constrained by an initial expression. Basically it
   --  transforms an unconstrained subtype indication into a constrained one.
   --  The expression may also be transformed in certain cases in order to
   --  avoid multiple evaulation. In the static allocation case, the general
   --  scheme is :
   --     Val : T := Expr;
   --        is transformed into
   --     Val : Constrained_Subtype_of_T := Maybe_Modified_Expr;
   --
   --  Here are the main cases :
   --
   --  <if Expr is a Slice>
   --    Val : T ([Index_Subtype (Expr)]) := Expr;
   --
   --  <elsif Expr is a String Literal>
   --    Val : T (T'First .. T'First + Length (string literal) - 1) := Expr;
   --
   --  <elsif Expr is Constrained>
   --    Val : Type_Of_Expr := Expr;
   --
   --  <elsif Expr is an entity_name>
   --    Val : T (contraints taken from Expr) := Expr;
   --
   --  <else>
   --    type Axxx is access all T;
   --    Rval : Axxx := Expr'ref;
   --    Val  : T (contraints taken from Rval) := Rval.all;
   --    ??? note: when the Expression is allocated in the secondary stack
   --              we could use it directly instead of copying it by declaring
   --              Val : T (...) renames Rval.all

   procedure Expand_Subtype_From_Expr
     (N             : Node_Id;
      Unc_Type      : Entity_Id;
      Subtype_Indic : Node_Id;
      Exp           : Node_Id)
   is
      Loc           : constant Source_Ptr := Sloc (N);
      Exp_Typ       : constant Entity_Id  := Etype (Exp);

   begin
      if not Expander_Active then
         return;
      end if;

      if Nkind (Exp) = N_Slice then
         Rewrite_Substitute_Tree (Subtype_Indic,
           Make_Subtype_Indication (Loc,
             Subtype_Mark => New_Reference_To (Unc_Type, Loc),
             Constraint =>
               Make_Index_Or_Discriminant_Constraint (Loc,
                 Constraints => New_List
                   (New_Reference_To (Etype (First_Index (Exp_Typ)), Loc)))));

      elsif Ekind (Exp_Typ) = E_String_Literal_Subtype then
         Rewrite_Substitute_Tree (Subtype_Indic,
           Make_Subtype_Indication (Loc,
             Subtype_Mark => New_Reference_To (Unc_Type, Loc),
             Constraint =>
               Make_Index_Or_Discriminant_Constraint (Loc,
                 Constraints => New_List (
                   Make_Literal_Range (Loc,
                     Literal_Typ => Exp_Typ,
                     Index_Typ   => Etype (First_Index (Unc_Type)))))));

      elsif Is_Constrained (Exp_Typ)
        and then not Is_Class_Wide_Type (Unc_Type)
      then
         Rewrite_Substitute_Tree (Subtype_Indic,
           New_Reference_To (Exp_Typ, Loc));

      else
         Remove_Side_Effects (Exp);
         Rewrite_Substitute_Tree (Subtype_Indic,
           Make_Subtype_From_Expr (Exp, Unc_Type));
      end if;
   end Expand_Subtype_From_Expr;

   ------------------
   -- Find_Prim_Op --
   ------------------

   function Find_Prim_Op (T : Entity_Id; Name : Name_Id) return Entity_Id is
      Prim : Elmt_Id;
      Typ  : Entity_Id := T;

   begin
      if Is_Class_Wide_Type (Typ) then
         Typ := Root_Type (Typ);
      end if;

      Typ := Underlying_Type (Typ);
      Prim := First_Elmt (Primitive_Operations (Typ));
      while Chars (Node (Prim)) /= Name loop
         Prim := Next_Elmt (Prim);
         pragma Assert (Present (Prim));
      end loop;

      return Node (Prim);
   end Find_Prim_Op;

   --------------------------------
   -- Make_Constraints_From_Expr --
   --------------------------------

   --  1. if Expr is an uncontrained array expression, creates
   --    Unc_Type(Expr'first(1)..Expr'Last(1),..., Expr'first(n)..Expr'last(n))

   --  2. if Expr is a unconstrained discriminated type expression, creates
   --    Unc_Type(Expr.Discr1, ... , Expr.Discr_n)

   --  3. if Expr is class-wide, creates an implicit class wide subtype

   function Make_Subtype_From_Expr
     (E       : Node_Id;
      Unc_Typ : Entity_Id)
      return    Node_Id
   is
      Loc         : constant Source_Ptr := Sloc (E);
      List_Constr : List_Id := New_List;
      D           : Entity_Id;

   begin
      if Is_Array_Type (Unc_Typ) then
         for J in 1 .. Number_Dimensions (Unc_Typ) loop
            Append_To (List_Constr,
              Make_Range (Loc,
                Low_Bound =>
                  Make_Attribute_Reference (Loc,
                    Prefix => Duplicate_Subexpr (E),
                    Attribute_Name => Name_First,
                    Expressions => New_List (
                      Make_Integer_Literal (Loc, Intval => UI_From_Int (J)))),
                High_Bound =>
                  Make_Attribute_Reference (Loc,
                    Prefix         => Duplicate_Subexpr (E),
                    Attribute_Name => Name_Last,
                    Expressions    => New_List (
                      Make_Integer_Literal (Loc,
                        Intval => UI_From_Int (J))))));
         end loop;

      elsif Is_Class_Wide_Type (Unc_Typ) then
         declare
            CW_Subtype : Entity_Id;
            EQ_Typ     : Entity_Id := Empty;

         begin
            if Expander_Active then
               EQ_Typ := Make_CW_Equivalent_Type (Unc_Typ, E);
            end if;

            CW_Subtype := New_Class_Wide_Subtype (Unc_Typ, E);
            Set_Equivalent_Type (CW_Subtype, EQ_Typ);

            return New_Occurrence_Of (CW_Subtype, Loc);
         end;

      else
         D := First_Discriminant (Unc_Typ);
         while (Present (D)) loop

            Append_To (List_Constr,
              Make_Selected_Component (Loc,
                Prefix        => Duplicate_Subexpr (E),
                Selector_Name => New_Reference_To (D, Loc)));

            D := Next_Discriminant (D);
         end loop;
      end if;

      return
        Make_Subtype_Indication (Loc,
          Subtype_Mark => New_Reference_To (Unc_Typ, Loc),
          Constraint   =>
            Make_Index_Or_Discriminant_Constraint (Loc,
              Constraints => List_Constr));
   end Make_Subtype_From_Expr;

   ------------------------
   -- Make_Literal_Range --
   ------------------------

   function Make_Literal_Range
     (Loc         : Source_Ptr;
      Literal_Typ : Entity_Id;
      Index_Typ   : Entity_Id)
      return        Node_Id
   is
   begin
         return
           Make_Range (Loc,
             Low_Bound =>
               Make_Attribute_Reference (Loc,
                 Prefix => New_Occurrence_Of (Index_Typ, Loc),
                 Attribute_Name => Name_First),
             High_Bound =>
               Make_Op_Subtract (Loc,
                  Left_Opnd =>
                    Make_Op_Add (Loc,
                      Left_Opnd =>
                        Make_Attribute_Reference (Loc,
                          Prefix => New_Occurrence_Of (Index_Typ, Loc),
                          Attribute_Name => Name_First),
                      Right_Opnd => Make_Integer_Literal (Loc,
                        String_Literal_Length (Literal_Typ))),
                  Right_Opnd => Make_Integer_Literal (Loc, Uint_1)));
   end Make_Literal_Range;

   ----------------------
   -- Make_Tagged_Copy --
   ----------------------

   --  Generate :
   --
   --      Annn : Tag := Lhs'tag;
   --      Bnnn : Finalizable_Ptr := Finalizable!(lhs).Prev;
   --      Cnnn : Finalizable_Ptr := Finalizable!(lhs).Next;
   --
   --      lhs := rhs;
   --
   --      lhs._tag := Annn;
   --      Finalizable!(lhs).Prev := Bnnn;
   --      Finalizable!(lhs).Next := Cnnn;

   procedure Expand_Tagged_Copy
     (N   : Node_Id;
      Lhs : Multi_Use.Exp_Id;
      Rhs : Multi_Use.Exp_Id;
      Typ : Entity_Id)
   is
      Loc      : constant Source_Ptr := Sloc (N);
      A        : constant Node_Id
              := Make_Defining_Identifier (Loc, New_Internal_Name ('A'));
      B        : constant Node_Id
                   := Make_Defining_Identifier (Loc, New_Internal_Name ('B'));
      C        : constant Node_Id
                   := Make_Defining_Identifier (Loc, New_Internal_Name ('C'));
      L        : List_Id;
      Ctrl_Ref : Node_Id;

   begin
      --  First save the tags and the finalization pointers

      L := New_List;

      if Is_Tagged_Type (Typ) then
         Append_To (L,
           Make_Object_Declaration (Loc,
             Defining_Identifier => A,
             Object_Definition => New_Reference_To (RTE (RE_Tag), Loc),
             Expression =>
               Make_Selected_Component (Loc,
                 Prefix        => Multi_Use.New_Ref (Lhs),
                 Selector_Name =>
                   New_Reference_To (Tag_Component (Typ), Loc))));
      end if;

      if Controlled_Type (Typ) then

         Ctrl_Ref := Multi_Use.New_Ref (Lhs);
         if Has_Controlled (Typ) then
            Ctrl_Ref :=
              Make_Selected_Component (Loc,
                Prefix => Ctrl_Ref,
                Selector_Name =>
                  New_Reference_To (Controller_Component (Typ), Loc));
         end if;

         Append_To (L,
           Make_Object_Declaration (Loc,
             Defining_Identifier => B,
             Object_Definition =>
               New_Reference_To (RTE (RE_Finalizable_Ptr), Loc),
             Expression =>
               Make_Selected_Component (Loc,
                 Prefix =>
                   Make_Unchecked_Type_Conversion (Loc,
                     Subtype_Mark =>
                       New_Reference_To (RTE (RE_Finalizable), Loc),
                     Expression => Ctrl_Ref),
                 Selector_Name => Make_Identifier (Loc, Name_Prev))));

         Append_To (L,
           Make_Object_Declaration (Loc,
             Defining_Identifier => C,
             Object_Definition =>
               New_Reference_To (RTE (RE_Finalizable_Ptr), Loc),
             Expression =>
               Make_Selected_Component (Loc,
                 Prefix =>
                   Make_Unchecked_Type_Conversion (Loc,
                     Subtype_Mark =>
                       New_Reference_To (RTE (RE_Finalizable), Loc),
                     Expression => New_Copy_Tree (Ctrl_Ref)),
                 Selector_Name => Make_Identifier (Loc, Name_Next))));
      end if;

      Insert_List_Before_And_Analyze (N, L);

      --  Do the copy

      Rewrite_Substitute_Tree (Name (N), Multi_Use.New_Ref (Lhs));
      Analyze (Name (N));

      Rewrite_Substitute_Tree (Expression (N), Multi_Use.New_Ref (Rhs));
      Analyze (Expression (N));

      --  Then restore tag and the finalization pointers

      L := New_List;

      if Is_Tagged_Type (Typ) then
         Append_To (L,
           Make_Assignment_Statement (Loc,
             Name =>
               Make_Selected_Component (Loc,
                 Prefix => Multi_Use.New_Ref (Lhs),
                 Selector_Name => New_Reference_To (Tag_Component (Typ), Loc)),
             Expression => New_Reference_To (A, Loc)));
      end if;

      if Controlled_Type (Typ) then

         Append_To (L,
           Make_Assignment_Statement (Loc,
             Name =>
               Make_Selected_Component (Loc,
                 Prefix =>
                   Make_Unchecked_Type_Conversion (Loc,
                     Subtype_Mark =>
                       New_Reference_To (RTE (RE_Finalizable), Loc),
                     Expression => New_Copy_Tree (Ctrl_Ref)),
                 Selector_Name => Make_Identifier (Loc, Name_Prev)),
             Expression => New_Reference_To (B, Loc)));

         Append_To (L,
           Make_Assignment_Statement (Loc,
             Name =>
               Make_Selected_Component (Loc,
                 Prefix =>
                   Make_Unchecked_Type_Conversion (Loc,
                     Subtype_Mark =>
                       New_Reference_To (RTE (RE_Finalizable), Loc),
                     Expression => New_Copy_Tree (Ctrl_Ref)),
                 Selector_Name => Make_Identifier (Loc, Name_Next)),
             Expression => New_Reference_To (C, Loc)));
      end if;

      Insert_List_After (N, L);
   end Expand_Tagged_Copy;

   -------------------------
   -- Remove_Side_Effects --
   -------------------------

   procedure Remove_Side_Effects
     (Exp      : Node_Id;
      Name_Req : Boolean := False)
   is
      Loc          : constant Source_Ptr := Sloc (Exp);
      Exp_Type     : constant Entity_Id  := Etype (Exp);
      Kind         : constant Node_Kind  := Nkind (Exp);
      Def_Id       : Entity_Id;
      Ref_Type     : Entity_Id;
      Res          : Node_Id;
      Ptr_Typ_Decl : Node_Id;
      New_Exp      : Node_Id;
      E            : Node_Id;
      Has_Prefix   : Boolean;

      function No_Range_Checks (L : List_Id) return Boolean;
      --  Determines if any of the expressions in L has a range check

      function Side_Effect_Free (N : Node_Id) return Boolean;
      --  Determines if the tree N represents an expession that is known
      --  not to have side effects, and for which no processing is required.

      function Side_Effect_Free (L : List_Id) return Boolean;
      --  Determines if all elements of the list L are side effect free

      function No_Range_Checks (L : List_Id) return Boolean is
         N : Node_Id;

      begin
         N := First (L);
         while Present (N) loop
            if Do_Range_Check (N) then
               return False;
            else
               N := Next (N);
            end if;
         end loop;

         return True;
      end No_Range_Checks;

      function Side_Effect_Free (N : Node_Id) return Boolean is
         K : constant Node_Kind := Nkind (N);

      begin
         --  Note on checks that could raise Constraint_Error. Strictly,
         --  if we take advantage of 11.6, these checks do not count as
         --  side effects. However, we would just as soon consider that
         --  they are side effects, since the backend CSE does not work
         --  very well on expressions which can raise Constraint_Error.

         --  An attribute reference is side effect free if it does not have
         --  an overflow check, and its expressions are side effect free.

         if K = N_Attribute_Reference then
            return not Do_Overflow_Check (N)
              and then Side_Effect_Free (Expressions (N));

         --  An entity is side effect free unless it is a function call, or
         --  a reference to a volatile variable and Name_Req is False. If
         --  Name_Req is True then we can't help returning a name which
         --  effectively allows multiple references in any case.

         elsif Is_Entity_Name (N)
           and then (Ekind (Entity (N)) /= E_Function)
           and then (not Is_Volatile (Entity (N)) or else Name_Req)
         then
            return True;

         --  A static expression is always side effect free

         elsif Is_OK_Static_Expression (N) then
            return True;

         --  A type conversion is side effect free if there are no checks,
         --  and the expression to be converted is side effect free.

         elsif K = N_Type_Conversion then
            return not Do_Overflow_Check (N)
              and then not Do_Range_Check (Expression (N))
              and then Side_Effect_Free (Expression (N));

         --  An unchecked type conversion or type qualification is side
         --  effect free if its argument is side effect free.

         elsif K = N_Qualified_Expression
           or else K = N_Unchecked_Type_Conversion
         then
            return Side_Effect_Free (Expression (N));

         --  A unary operator is side effect free if there is no overflow
         --  check, and the operand is side effect free.

         elsif K in N_Unary_Op then
            return not Do_Overflow_Check (N)
              and then Side_Effect_Free (Right_Opnd (N));

         --  A binary operator is side effect free if there is no overflow
         --  check, and both operands are side effect free.

         elsif K in N_Binary_Op then
            return not Do_Overflow_Check (N)
              and then Side_Effect_Free (Left_Opnd  (N))
              and then Side_Effect_Free (Right_Opnd (N));

         --  An explicit dereference or selected component is side effect
         --  free if its prefix is side effect free and there is no access
         --  check. (??? for now, we do not check for access checks since
         --  the expanded code generate piles of those and is not abler yet
         --  to generate them with checks off)

         elsif K = N_Explicit_Dereference
           or else K = N_Selected_Component
         then
            return Side_Effect_Free (Prefix (N));
--              and then not Do_Access_Check (N);

         --  An indexed component can be copied if the prefix is copyable
         --  and all the indexing expressions are copyable and there is
         --  no access check and no range checks.

         elsif K = N_Indexed_Component then
            return not Do_Access_Check (N)
              and then Side_Effect_Free (Prefix (N))
              and then Side_Effect_Free (Expressions (N))
              and then No_Range_Checks (Expressions (N));

         --  We consider that anything else has side effects. This is a bit
         --  crude, but we are pretty close for most common cases, and we
         --  are certainly correct (i.e. we never return True when the
         --  answer should be False).

         else
            return False;
         end if;
      end Side_Effect_Free;

      function Side_Effect_Free (L : List_Id) return Boolean is
         N : Node_Id;

      begin
         if L = No_List or else L = Error_List then
            return True;

         else
            N := First (L);

            while Present (N) loop
               if not Side_Effect_Free (N) then
                  return False;
               else
                  N := Next (N);
               end if;
            end loop;

            return True;
         end if;
      end Side_Effect_Free;

   --  Start of processing for Remove_Side_Effects

   begin
      --  First we check for the cases which are side effect already

      if Side_Effect_Free (Exp) then
         return;

         --  If the expression has the form v.all then we can just capture
         --  the pointer, and then do an explicit dereference on the result.

      elsif Nkind (Exp) = N_Explicit_Dereference then
         Def_Id :=
           Make_Defining_Identifier (Loc, New_Internal_Name ('R'));
         Res :=
           Make_Explicit_Dereference (Loc, New_Reference_To (Def_Id, Loc));

         Insert_Action (Exp,
           Make_Object_Declaration (Loc,
             Defining_Identifier => Def_Id,
             Object_Definition   =>
               New_Reference_To (Etype (Prefix (Exp)), Loc),
             Expression          => Relocate_Node (Prefix (Exp))));

      --  For expressions that denote objects, we can use a renaming scheme
      --  We skip using this if we have a volatile variable and we do not
      --  have Nam_Req set true (see comments above for Side_Effect_Free).

      elsif Is_Object_Reference (Exp)
        and then (Name_Req
                   or else not Is_Entity_Name (Exp)
                   or else not Is_Volatile (Entity (Exp)))
      then
         Def_Id := Make_Defining_Identifier (Loc, New_Internal_Name ('R'));
         Res := New_Reference_To (Def_Id, Loc);

         Insert_Action (Exp,
           Make_Object_Renaming_Declaration (Loc,
             Defining_Identifier => Def_Id,
             Subtype_Mark        => New_Reference_To (Exp_Type, Loc),
             Name                => Relocate_Node (Exp)));

      --  If it is a scalar type, just make a copy

      elsif Is_Elementary_Type (Exp_Type) then
         Def_Id := Make_Defining_Identifier (Loc, New_Internal_Name ('R'));
         Res := New_Reference_To (Def_Id, Loc);

         Insert_Action (Exp,
           Make_Object_Declaration (Loc,
             Defining_Identifier => Def_Id,
             Object_Definition   => New_Reference_To (Exp_Type, Loc),
             Expression          => Relocate_Node (Exp)));

      --  Otherwise we generate a reference to the value

      else
         Ref_Type := Make_Defining_Identifier (Loc, New_Internal_Name ('A'));

         Ptr_Typ_Decl :=
           Make_Full_Type_Declaration (Loc,
             Defining_Identifier => Ref_Type,
             Type_Definition =>
               Make_Access_To_Object_Definition (Loc,
                 All_Present => True,
                 Subtype_Indication =>
                   New_Reference_To (Exp_Type, Loc)));

         --  If the type of the expression is an implicit type defined
         --  in the expression node itself it better be transfered to
         --  the access type declaration to avoid forward references.
         --  Same thing for expression actions appearing in the prefix

         Has_Prefix := False
           or else Kind = N_Attribute_Reference
           or else Kind = N_Expanded_Name
           or else Kind = N_Explicit_Dereference
           or else Kind = N_Indexed_Component
           or else Kind = N_Reference
           or else Kind = N_Selected_Component
           or else Kind = N_Slice;

         if Is_Itype (Exp_Type) then
            if Nkind (Exp) in N_Has_Itypes then
               Transfer_Itypes (From => Exp, To => Ptr_Typ_Decl);
            elsif Has_Prefix
              and then Nkind (Prefix (Exp)) in N_Has_Itypes
            then
               Transfer_Itypes (From => Prefix (Exp), To => Ptr_Typ_Decl);
            end if;
         end if;

         --  Furthermore if the expression is an expression-actions then
         --  move actions upfront too since the type can depend on those
         --  actions

         if Nkind (Exp) = N_Expression_Actions then
            Insert_Actions (Exp, Actions (Exp));
            E := Expression (Exp);

         elsif Has_Prefix
           and then Nkind (Prefix (Exp)) = N_Expression_Actions
         then
            Insert_Actions (Exp, Actions (Prefix (Exp)));
            Set_Prefix (Exp, Expression (Prefix (Exp)));
            E := Exp;

         else
            E := Exp;
         end if;

         Insert_Action (Exp, Ptr_Typ_Decl);

         Def_Id := Make_Defining_Identifier (Loc, New_Internal_Name ('R'));
         Res :=
           Make_Explicit_Dereference (Loc,
             Prefix => New_Reference_To (Def_Id, Loc));

         if Nkind (E) = N_Explicit_Dereference then
            New_Exp := Relocate_Node (Prefix (E));
         else
            New_Exp := Make_Reference (Loc, Relocate_Node (E));
         end if;

         Insert_Action (Exp,
           Make_Object_Declaration (Loc,
             Defining_Identifier => Def_Id,
             Object_Definition   => New_Reference_To (Ref_Type, Loc),
             Expression          => New_Exp));
      end if;

      --  Preserve the Assignment_OK flag in all copies, since at least
      --  one copy may be used in a context where this flag must be set
      --  (otherwise why would the flag be set in the first place).

      Set_Assignment_OK (Res, Assignment_OK (Exp));

      --  Finally rewrite the original expressoin and we are done

      Rewrite_Substitute_Tree (Exp, Res);
      Analyze (Exp);
      Resolve (Exp, Exp_Type);
   end Remove_Side_Effects;

   --------------------------
   -- Unchecked_Convert_To --
   --------------------------

   function Unchecked_Convert_To
     (Typ  : Entity_Id;
      Expr : Node_Id)
      return Node_Id
   is
   begin
      if Present (Etype (Expr))
        and then Base_Type (Etype (Expr)) = Typ
      then
         return Relocate_Node (Expr);
      else
         return
           Make_Unchecked_Type_Conversion (Sloc (Expr),
             Subtype_Mark => New_Occurrence_Of (Typ, Sloc (Expr)),
             Expression   => Relocate_Node (Expr));
      end if;
   end Unchecked_Convert_To;

   ----------------------------
   -- Wrap_Cleanup_Procedure --
   ----------------------------

   procedure Wrap_Cleanup_Procedure (N : Node_Id) is
      Loc   : constant Source_Ptr := Sloc (N);
      Stseq : constant Node_Id    := Handled_Statement_Sequence (N);
      Stmts : constant List_Id    := Statements (Stseq);

   begin
      Prepend_To (Stmts, Build_Runtime_Call (Loc, RE_Abort_Defer));
      Append_To  (Stmts, Build_Runtime_Call (Loc, RE_Abort_Undefer));
   end Wrap_Cleanup_Procedure;

end Exp_Util;
